from electricore.core.périmètre.modèles import HistoriquePérimètre, SituationPérimètre, ModificationContractuelleImpactante
from electricore.core.périmètre.fonctions import (
    extraire_situation,
    extraire_historique_à_date,
    extraire_période,
    extraite_relevés_entrées,
    extraite_relevés_sorties,
    extraire_modifications_impactantes,
    detecter_points_de_rupture,
    inserer_evenements_facturation,
    extraire_releves_evenements
)


from electricore.inputs.flux.modèles import FluxR151, FluxC15, FluxF1X
from electricore.inputs.flux.flux_relevés_Rxx import lire_flux_r151
from electricore.inputs.flux.flux_contrats_Cxx import lire_flux_c15
from electricore.inputs.flux.flux_factures_Fxx import lire_flux_f1x


import pandas as pd
import pandera.pandas as pa
from pandera.typing import DataFrame
from electricore.core.périmètre.modèles import HistoriquePérimètre
from electricore.inputs.flux.modèles import FluxC15

# Point sur les cas relevés Avant et/ou relevé Après. 
# Corrélations calculées sur les données dispo le 2025/02/26 :
# | Evenement_Declencheur | Avant | Après | ~Avant | ~Après |
# |-----------------------|-------|-------|--------|--------|
# | AUTRE                 | 0.0   | 0.0   | 1.0    | 1.0    |
# | CFNE                  | 0.0   | 0.95  | 1.0    | 0.05   |
# | CFNS                  | 1.0   | 0.0   | 0.0    | 1.0    |
# | CMAT                  | 0.11  | 0.78  | 0.89   | 0.22   |
# | MCF                   | 1.0   | 1.0   | 0.0    | 0.0    |
# | MCT                   | 0.67  | 1.0   | 0.33   | 0.0    |
# | MDACT                 | 0.0   | 0.0   | 1.0    | 1.0    |
# | MDPRM                 | 0.0   | 0.0   | 1.0    | 1.0    |
# | MES                   | 0.0   | 0.98  | 1.0    | 0.02   |
# | PMES                  | 0.04  | 0.88  | 0.96   | 0.12   |
# | RES                   | 1.0   | 0.0   | 0.0    | 1.0    |

# AUTRE, MDACT, MDPRM = jamais de relevés, ni Avant, ni Après
# MES, PMES, CFNE, jamais Avant, PRESQUE toujours Après (On a rien quand c'est nature = ESTIMÉ...)
# RES, CFNS = toujours Avant, jamais Après (makes sense)
# MCF toujours Avant ET Après (Modification de la programmation du calendrier fournisseur)
# MCT souvent Avant, toujours Après (Modification de la formule tarifaire d’acheminement ou de la puissance souscrite ou du statut d’Autoconsommation Collective)
# CMAT rarement Avant, souvent Après (Changement de compteur ou de disjoncteur ou Activation du calendrier Distributeur)
# Pas encore rencontré : MDBRA (Modification de données de branchement)

@pa.check_types()
def lire_flux_c15(source: pd.DataFrame) -> DataFrame[HistoriquePérimètre]:

    df: DataFrame[FluxC15] = FluxC15.validate(source)
    df["Source"] = "flux_C15"

    return HistoriquePérimètre.validate(df)


import pandas as pd
import pandera.pandas as pa
from pandera.typing import DataFrame

from electricore.core.relevés.modèles import RelevéIndex, RequêteRelevé

@pa.check_types
def interroger_relevés(
    requêtes: DataFrame[RequêteRelevé], 
    relevés: DataFrame[RelevéIndex], 
    tolérance: pd.Timedelta = pd.Timedelta(hours=4)
) -> DataFrame[RelevéIndex]:
    """
    🔍 Interroge les relevés d'index pour récupérer les index correspondant à une liste de dates et PDL,
    avec une tolérance d'une heure sur la Date_Releve.

    Args:
        requêtes (DataFrame[RequêteRelevé]): DataFrame contenant les colonnes "Date_Releve" et "pdl" pour la requête.
        relevés (DataFrame[RelevéIndex]): DataFrame contenant les relevés d'index.

    Returns:
        DataFrame: DataFrame contenant les relevés correspondant aux requêtes.
    """

    # Sauvegarde de l'index d'origine
    requêtes_avec_index_col = requêtes.copy().reset_index()

    # 🔄 Jointure avec tolérance
    relevés_proches = pd.merge_asof(
        requêtes_avec_index_col.sort_values(by=["Date_Releve"]), 
        relevés.copy().sort_values(by=["Date_Releve"]), 
        on="Date_Releve", 
        by="pdl", 
        direction="nearest",
        tolerance=tolérance 
    )

    return relevés_proches.dropna(subset=['Source']).set_index('index')



from electricore.core.périmètre import HistoriquePérimètre
from electricore.core.abonnements.modèles import PeriodeAbonnement

import pandera.pandas as pa
from pandera.typing import DataFrame

import pandas as pd

from babel.dates import format_date

@pa.check_types
def generer_periodes_abonnement(historique: DataFrame[HistoriquePérimètre]) -> DataFrame[PeriodeAbonnement]:
    """
    Génère les périodes homogènes d'abonnement à partir des événements impactant le TURPE fixe.
    """
    # 1. Filtrer les événements pertinents
    filtres = (
        (historique["impact_turpe_fixe"] == True) &
        (historique["Ref_Situation_Contractuelle"].notna())
    )
    abonnements = historique[filtres].copy()

    # 2. Trier par ref et date
    abonnements = abonnements.sort_values(["Ref_Situation_Contractuelle", "Date_Evenement"])

    # 3. Construire les débuts et fins de période
    abonnements["periode_debut"] = abonnements["Date_Evenement"]
    abonnements["periode_fin"] = abonnements.groupby("Ref_Situation_Contractuelle")["Date_Evenement"].shift(-1)

    # 4. Ne garder que les lignes valides
    periodes = abonnements.dropna(subset=["periode_fin"]).copy()

    # 5. Ajouter FTA, Puissance, et nb jours (arrondi à la journée, pas de time)
    periodes["Formule_Tarifaire_Acheminement"] = periodes["Formule_Tarifaire_Acheminement"]
    periodes["Puissance_Souscrite"] = periodes["Puissance_Souscrite"]
    periodes["nb_jours"] = (periodes["periode_fin"].dt.normalize() - periodes["periode_debut"].dt.normalize()).dt.days

    # 5. Ajouter FTA, Puissance, et nb jours (arrondi à la journée, pas de time)
    periodes["Formule_Tarifaire_Acheminement"] = periodes["Formule_Tarifaire_Acheminement"]
    periodes["Puissance_Souscrite"] = periodes["Puissance_Souscrite"]
    periodes["nb_jours"] = (periodes["periode_fin"].dt.normalize() - periodes["periode_debut"].dt.normalize()).dt.days

    # 6. Ajout lisibles
    periodes['periode_debut_lisible'] = periodes['periode_debut'].apply(
        lambda d: format_date(d, "d MMMM yyyy", locale="fr_FR") if not pd.isna(d) else None
    )
    periodes['periode_fin_lisible'] = periodes['periode_fin'].apply(
        lambda d: format_date(d, "d MMMM yyyy", locale="fr_FR") if not pd.isna(d) else "en cours"
    )

    periodes['mois_annee'] = periodes['periode_debut'].apply(
        lambda d: format_date(d, "LLLL yyyy", locale="fr_FR")
    )
    return periodes[[
        "Ref_Situation_Contractuelle",
        "mois_annee",
        "periode_debut_lisible",
        "periode_fin_lisible",
        "Formule_Tarifaire_Acheminement",
        "Puissance_Souscrite",
        "nb_jours",
        "periode_debut",
    ]].reset_index(drop=True)


import pandas as pd
from pathlib import Path

from zoneinfo import ZoneInfo

import pandera.pandas as pa
from pandera.typing import DataFrame
from electricore.core.abonnements.modèles import PeriodeAbonnement

from icecream import ic
PARIS_TZ = ZoneInfo("Europe/Paris")

def load_turpe_rules() -> pd.DataFrame:
    """
    Charge les règles TURPE à partir du fichier CSV.
    """
    file_path = Path(__file__).parent / "turpe_rules.csv"
    turpe_rules = pd.read_csv(file_path, parse_dates=["start", "end"])

    # Convertir en date avec fuseau horaire
    
    turpe_rules["start"] = pd.to_datetime(turpe_rules["start"]).dt.tz_localize(PARIS_TZ)
    turpe_rules["end"] = pd.to_datetime(turpe_rules["end"]).dt.tz_localize(PARIS_TZ, ambiguous='NaT')

    # Convertir toutes les colonnes non-meta en float
    meta_columns = ["start", "end", "Formule_Tarifaire_Acheminement"]
    numeric_columns = [col for col in turpe_rules.columns if col not in meta_columns]
    turpe_rules[numeric_columns] = turpe_rules[numeric_columns].astype(float)
    return turpe_rules

def get_applicable_rules(start: pd.Timestamp, end: pd.Timestamp, rules: pd.DataFrame|None=None) -> pd.DataFrame:
    """
    Retourne les règles TURPE applicables pour une période donnée.

    :param start: Date de début de la période en pd.Timestamp.
    :param end: Date de fin de la période en pd.Timestamp.
    :param rules_df: DataFrame contenant les colonnes [Formule_Tarifaire_Acheminement, start, end, b, HPH, HCH, ...].
    :return: DataFrame avec les règles applicables.
    """
    if rules is None:
        rules = load_turpe_rules()

    # Gérer les valeurs NaT dans la colonne "end"
    rules["end"] = rules["end"].apply(lambda x: x if pd.notna(x) else pd.Timestamp.max.tz_localize(start.tz))

    # Filtrer les règles applicables à la période donnée
    applicable_rules = rules[
        (rules["start"] < end) &
        (rules["end"] > start)
    ]

    return applicable_rules

def compute_turpe(entries: pd.DataFrame, rules: pd.DataFrame) -> pd.DataFrame:
    """
    Calcule le TURPE pour chaque entrée en utilisant les règles données.

    :param entries_df: DataFrame contenant les colonnes [start, end, FTA, Puissance_Souscrite, HPH, HCH, ...].
    :param rules_df: DataFrame contenant les colonnes [FTA, start, end, b, HPH, HCH, ...].
    :return: DataFrame avec les coûts calculés.
    """
    # Vérifier les doublons dans la colonne FTA
    duplicated = rules[rules.duplicated(subset=['Formule_Tarifaire_Acheminement'], keep=False)]
    if not duplicated.empty:
        raise ValueError(f"Doublons détectés dans la colonne Formule_Tarifaire_Acheminement :\n{duplicated}")

    # Fusionner les règles avec les entrées sur la clé FTA
    merged = pd.merge(entries, rules, on="Formule_Tarifaire_Acheminement", suffixes=("_entry", "_rule"))
    
    conso_cols = ["HPH", "HCH", "HPB", "HCB", "HP", "HC", "BASE"]
    # Calcul vectoriel des coûts fixes et variables
    merged["CS_fixe"] = merged["b"] * merged["Puissance_Souscrite"]
    for col in conso_cols:
        merged[f"turpe_{col}"] =  pd.to_numeric(merged[f"{col}_entry"] * merged[f"{col}_rule"] / 100, errors='coerce')

    merged["turpe_fixe_annuel"] = merged["CS_fixe"] + merged["cg"] + merged["cc"]
    merged["turpe_fixe_j"] = merged["turpe_fixe_annuel"] / 365

    merged["turpe_fixe"] = merged["turpe_fixe_j"] * merged["j"]

    merged["turpe_var"] = merged[['turpe_'+col for col in conso_cols]].sum(axis=1, min_count=1)
    merged["turpe"] = merged["turpe_fixe"] + merged["turpe_var"]
    
    columns_to_rename = {'start': 'Version_Turpe'} | {c+'_entry': c for c in conso_cols}
    merged = merged.rename(columns=columns_to_rename)
    
    columns_to_drop = [col for col in merged.columns if col.endswith('_entry')]+['end']

    merged = merged.drop(columns=columns_to_drop)

    merged['Version_Turpe'] = merged['Version_Turpe'].dt.date
    return merged.round(2)

@pa.check_types
def ajouter_turpe_fixe(
    periodes: DataFrame[PeriodeAbonnement],
    regles: pd.DataFrame
) -> DataFrame[PeriodeAbonnement]:
    """
    Calcule le TURPE fixe pour chaque période d'abonnement, selon les règles tarifaires.

    Args:
        periodes: Périodes d'abonnement homogènes (Formule_Tarifaire_Acheminement, Puissance_Souscrite, nb_jours)
        regles: Règles tarifaires TURPE, avec colonnes : FTA, b, cg, cc

    Returns:
        Périodes enrichies avec `turpe_fixe_journalier` et `turpe_fixe`
    """

    # Réplication croisée FTA entre périodes et règles
    df = pd.merge(
        periodes,
        regles,
        on="Formule_Tarifaire_Acheminement",
        how="left",
        suffixes=("", "_regle")
    )

    # Filtrage temporel
    df["end"] = df["end"].fillna(pd.Timestamp("2100-01-01").tz_localize(PARIS_TZ))

    mask = (df["periode_debut"] >= df["start"]) & (df["periode_debut"] < df["end"])
    df = df[mask].copy()

    # Calculs
    df["turpe_fixe_annuel"] = (df["b"] * df["Puissance_Souscrite"]) + df["cg"] + df["cc"]
    df["turpe_fixe_journalier"] = df["turpe_fixe_annuel"] / 365
    df["turpe_fixe"] = (df["turpe_fixe_journalier"] * df["nb_jours"]).round(2)

    return df[PeriodeAbonnement.to_schema().columns.keys()]
    



import pandera.pandas as pa
import pandas as pd
import numpy as np

from pandera.typing import DataFrame
from electricore.core.périmètre import (
    HistoriquePérimètre, SituationPérimètre, ModificationContractuelleImpactante,
    extraire_situation, extraire_période,
    extraite_relevés_entrées, extraite_relevés_sorties
)
from electricore.core.relevés import RelevéIndex, interroger_relevés
from electricore.core.énergies.modèles import BaseCalculEnergies

from icecream import ic

def préparer_base_énergies(
    historique: DataFrame[HistoriquePérimètre], deb: pd.Timestamp, fin: pd.Timestamp
) -> DataFrame[BaseCalculEnergies]:
    """
    🏗️ Prépare la base des énergies en identifiant les entrées, sorties et MCT dans la période.

    Args:
        historique (DataFrame[HistoriquePérimètre]): Historique des situations contractuelles.
        deb (pd.Timestamp): Début de la période de calcul des énergies.
        fin (pd.Timestamp): Fin de la période de calcul des énergies.

    Returns:
        DataFrame[SituationPérimètre]: Situation contractuelle enrichie pour le calcul des énergies.
    """
    colonnes_meta_releve = ['Unité', 'Précision', 'Source']
    colonnes_releve = ['Id_Calendrier_Distributeur', 'Date_Releve', 'Nature_Index', 'HP', 'HC', 'HCH', 'HPH', 'HPB', 'HCB', 'BASE']



    # 1) On récupére la situation du périmètre telle qu'elle était à la date de fin
    situation = extraire_situation(fin, historique)

    # 2) On filtre pour n'avoir que les PDLs en service, ou dont le service c'est terminé dans la période.
    # (car pour les autres, aka terminés avant la période, il n'y a rien a calculer pour la période)
    _masque = (situation['Etat_Contractuel'] == 'EN SERVICE') | (
        (situation['Etat_Contractuel'] == 'RESILIE') & (situation['Date_Evenement'] >= deb)
    )
    # Ajouter ici des colonnes supp si besoin de l'info plus loin
    colonnes_évenement = ['Ref_Situation_Contractuelle', 
                          'pdl', 
                          'Formule_Tarifaire_Acheminement', 
                          'Puissance_Souscrite',
                          'Type_Compteur', 'Num_Compteur', 'Num_Depannage']
    base = (
        situation[_masque]
        .drop(columns=[col for col in situation if col not in colonnes_évenement])
        .sort_values(by="Ref_Situation_Contractuelle")
        .copy()
    )

    # 3) On interroge le périmètre sur les éventuelles entrées et sorties, et on récupére les relevés d'index associés.
    période: DataFrame[HistoriquePérimètre] = extraire_période(deb, fin, historique)

    entrées: DataFrame[RelevéIndex] = (
        extraite_relevés_entrées(période)
        .set_index('Ref_Situation_Contractuelle')
        .drop(columns=['pdl'])
        .add_suffix('_deb')
        .assign(Entree=True)
    )
    sorties: DataFrame[RelevéIndex] = (
        extraite_relevés_sorties(période)
        .set_index('Ref_Situation_Contractuelle')
        .drop(columns=['pdl'])
        .add_suffix('_fin')
        .assign(Sortie=True)
    )

    # On les fusionne dans la base
    base = (
        base
        .merge(entrées, how='left', left_on='Ref_Situation_Contractuelle', right_index=True)
        .merge(sorties, how='left', left_on='Ref_Situation_Contractuelle', right_index=True)
        .fillna({'Entree': False, 'Sortie': False})
        .infer_objects(copy=False)  # Explicitly infer proper dtypes after fillna
    )

    return base

# @pa.check_types
def découper_périodes(
    base_énergies: DataFrame[BaseCalculEnergies],
    modifications: DataFrame[ModificationContractuelleImpactante]
) -> DataFrame[BaseCalculEnergies]:
    """
    📌 Découpe la base de calcul d'énergies en sous-périodes calculables.

    Cette fonction segmente les périodes impactées par des modifications contractuelles (MCT)
    en sous-périodes homogènes, prêtes pour les calculs d’énergies.

    - Ajoute des points de découpage à chaque MCT.
    - Génère des périodes couvrantes et calculables avec les valeurs mises à jour.

    🚀 Résultat : Des périodes propres et exploitables pour le calcul des énergies.
    """

    # 1️⃣ **Séparer les périodes impactées et non impactées**
    impactées = base_énergies[
        base_énergies["Ref_Situation_Contractuelle"].isin(modifications["Ref_Situation_Contractuelle"])
    ]
    non_impactées = base_énergies[
        ~base_énergies["Ref_Situation_Contractuelle"].isin(modifications["Ref_Situation_Contractuelle"])
    ]

    # 2️⃣ **Générer les sous-périodes pour les lignes impactées**
    all_periods = []

    for ref_situation, modifs in modifications.groupby("Ref_Situation_Contractuelle"):
        # Trier les modifications chronologiquement
        modifs = modifs.sort_values(by="Date_Evenement")

        # Récupérer la ligne initiale
        base_ligne = impactées[impactées["Ref_Situation_Contractuelle"] == ref_situation].iloc[0].copy()

        # Initialisation des dates de découpage
        dates_coupure = [base_ligne["Date_Releve_deb"]] + \
                        modifs["Date_Evenement"].tolist() + \
                        [base_ligne["Date_Releve_fin"]]
        dates_coupure = sorted(set(dates_coupure))

        # 3️⃣ **Créer une ligne par sous-période**
        for i in range(len(dates_coupure) - 1):
            periode = base_ligne.copy()
            periode["Date_Début"] = dates_coupure[i]
            periode["Date_Fin"] = dates_coupure[i + 1]

            # Appliquer la modification contractuelle si elle intervient à cette date
            modif_courante = modifs[modifs["Date_Evenement"] == dates_coupure[i]]
            if not modif_courante.empty:
                modif_courante = modif_courante.iloc[0]
                periode["Puissance_Souscrite"] = modif_courante["Avant_Puissance_Souscrite"]
                periode["Formule_Tarifaire_Acheminement"] = modif_courante["Avant_Formule_Tarifaire_Acheminement"]

            all_periods.append(periode)
    return all_periods

    # 4️⃣ **Concaténer les périodes impactées + les non impactées**
    base_decoupée = pd.concat([non_impactées] + all_periods, ignore_index=True)

    return base_decoupée

def ajouter_relevés(
    base: DataFrame[BaseCalculEnergies], 
    relevés: DataFrame[RelevéIndex],
    suffixe: str = "_deb"  # Valeur par défaut "_deb", peut être "_fin"
) -> DataFrame[BaseCalculEnergies]:
    """
    🔄 Ajoute les relevés manquants dans la base de calcul des énergies.

    Args:
        base (DataFrame[BaseCalculEnergies]): Base existante des calculs d'énergie.
        relevés (DataFrame[RelevéIndex]): Relevés d'index disponibles.
        suffixe (str, optional): Suffixe qui identifie s'il s'agit de relevés de début ("_deb") 
                                ou de fin ("_fin"). Par défaut "_deb".

    Returns:
        DataFrame[BaseCalculEnergies]: Base mise à jour avec les relevés ajoutés.
    """
    # Dynamiquement construire les noms de colonnes basés sur le suffixe
    col_date_releve = f"Date_Releve{suffixe}"
    col_source = f"Source{suffixe}"
    
    # 🏷️ Extraire les paires (Date_Releve, pdl) manquantes dans la base
    requêtes_manquantes = (
        base
        .loc[base[col_source].isna(), [col_date_releve, "pdl"]]
        .rename(columns={col_date_releve: 'Date_Releve'})
        .drop_duplicates()
    )
    if requêtes_manquantes.empty:
        return base  # ✅ Rien à ajouter, on retourne la base inchangée.
    
    # 🔍 Récupération des relevés manquants
    relevés_trouvés = (
        interroger_relevés(requêtes_manquantes, relevés)
        .add_suffix(suffixe)
        .rename(columns={f'pdl{suffixe}': 'pdl'})
    )
    
    # Préparation pour la mise à jour
    base_mise_a_jour = base.copy()
    
    # Mise à jour
    base_mise_a_jour.update(relevés_trouvés)

    return base_mise_a_jour

def calculer_energies(
    base: DataFrame[BaseCalculEnergies],
    inclure_jour_fin: bool=False
) -> DataFrame[BaseCalculEnergies]:
    """
    ⚡ Calcule les énergies consommées en faisant la différence entre les index de fin et de début
    pour les lignes où les calendriers de distribution sont identiques.

    Args:
        base (DataFrame[BaseCalculEnergies]): Base contenant les relevés de début et de fin.

    Returns:
        DataFrame[BaseCalculEnergies]: Base avec les énergies calculées.
    """
    # Liste des cadrans d'index à traiter
    cadrans = ['HPH', 'HPB', 'HCH', 'HCB', 'HP', 'HC', 'BASE']
    
    # Copie de la base pour ne pas modifier l'original
    resultat = base.copy()
    
    # Vérification de l'égalité des calendriers distributeur
    calendriers_identiques = (
        resultat["Id_Calendrier_Distributeur_deb"] == 
        resultat["Id_Calendrier_Distributeur_fin"]
    )
    
    # On ne calcule les énergies que pour les lignes où les calendriers sont identiques
    lignes_valides = resultat[calendriers_identiques].index
    
    if len(lignes_valides) == 0:
        print("⚠️ Aucune ligne avec des calendriers identiques trouvée.")
        return resultat
    
    # Pour chaque cadran, calculer l'énergie consommée
    for cadran in cadrans:
        col_deb = f"{cadran}_deb"
        col_fin = f"{cadran}_fin"
        col_energie = cadran
        
        # Calculer l'énergie comme la différence entre l'index de fin et de début
        # On arrondit à l'entier inférieur pour éviter les problèmes de précision différentes entre les relevés,
        resultat.loc[lignes_valides, col_energie] = (
            np.floor(resultat.loc[lignes_valides, col_fin]) - 
            np.floor(resultat.loc[lignes_valides, col_deb])
        )
        
        # Vérifier les valeurs négatives (anomalies potentielles)
        nb_negatifs = (resultat.loc[lignes_valides, col_energie] < 0).sum()
        if nb_negatifs > 0:
            print(f"⚠️ {nb_negatifs} valeurs négatives détectées pour {col_energie}")
    
    # Ajouter une colonne pour indiquer si l'énergie a été calculée
    resultat["Energie_Calculee"] = False
    resultat.loc[lignes_valides, "Energie_Calculee"] = True
    
    # Calculer la somme totale des énergies (tous cadrans confondus)
        # Calcul du nombre de jours entre les deux relevés
    resultat['j'] = (
        resultat["Date_Releve_fin"].dt.date - resultat["Date_Releve_deb"].dt.date
    ).apply(lambda x: x.days + (1 if inclure_jour_fin else 0))

    # Calculer HP et HC en prenant la somme des colonnes correspondantes
    resultat['HP'] = resultat[['HPH', 'HPB', 'HP']].sum(axis=1, min_count=1)
    resultat['HC'] = resultat[['HCH', 'HCB', 'HC']].sum(axis=1, min_count=1)

    # Calculer BASE uniquement là où BASE est NaN
    resultat.loc[resultat['BASE'].isna(), 'BASE'] = resultat[['HP', 'HC']].sum(axis=1, min_count=1)
    
    return resultat


import pandas as pd
import pandera.pandas as pa
from pandera.typing import DataFrame
from electricore.core.relevés.modèles import RelevéIndex
from electricore.inputs.flux.modèles import FluxR151

# Flux R151 énergies quotidiennes
@pa.check_types()
def lire_flux_r151(source: DataFrame[FluxR151]) -> DataFrame[RelevéIndex]:

    df: DataFrame[FluxR151] = source.copy()
    df["Source"] = "flux_R151"

    # Supprimer des lignes ou 'Id_Calendrier_Distributeur' == 'INCONNU'
    # Dans la doc enedis, c'est quand il n'y a pas de mesure. Pourquoi en faire une entrée alors ? Yo no sé
    # Supprimer les lignes où 'Id_Calendrier_Distributeur' == 'INCONNU' et == 'DN999999'
    # Note : Je ne sais pas ce que DN999999 vient faire là. C'est dans ces lignes qu'on retrouve des valeurs dans la colonnes INCONNU. Pour l'instant on supprime.   
    df = df[~df['Id_Calendrier_Distributeur'].isin(['INCONNU', 'DN999999'])]
    
    # Réordonner les colonnes pour correspondre au modèle RelevéIndex attendu
    # ordre_colonnes = RelevéIndex.to_schema().columns.keys()
    # df = df[ordre_colonnes]

    # Supprimer des colonnes si présentes
    _to_drop: list[str] = [c for c in ['INCONNU'] if c in df.columns]
    df = df.drop(columns=_to_drop)

    return df


import pandas as pd

# def energies_et_taxes(deb: pd.Timestamp, fin: pd.Timestamp, base: pd.DataFrame) -> pd.DataFrame:
#     """
#     Calcule les énergies et les taxes pour une période donnée, sur l'ensemble du périmètre
#     """
    
#     complet = ajout_dates_par_defaut(deb, fin, base)
#     energies = calcul_energie(complet)
#     energies['Puissance_Souscrite'] = pd.to_numeric(energies['Puissance_Souscrite'])
#     rules = get_applicable_rules(deb, fin)
#     turpe = compute_turpe(entries=energies, rules=rules)
#     final = validation(
#         supprimer_colonnes(
#         fusion_des_sous_periode(turpe)))
#     return final.round(2)
from icecream import ic
import pandera.pandas as pa
from pandera.typing import DataFrame

from electricore.inputs.flux import (
    FluxC15, FluxR151, 
    lire_flux_c15, lire_flux_r151
)
from electricore.core.périmètre import (
    HistoriquePérimètre,
    ModificationContractuelleImpactante,
    extraire_historique_à_date,
    extraire_période,
    extraire_modifications_impactantes
)
from electricore.core.relevés import (
    RelevéIndex, 
)
from electricore.core.énergies.fonctions import (
    BaseCalculEnergies,
    préparer_base_énergies, 
    ajouter_relevés, 
    calculer_energies
)
from electricore.core.taxes.turpe import (
    get_applicable_rules,
    compute_turpe,
)
# TODO rename facturation depuis flux ou un truc du genre. 
def facturation_flux(deb: pd.Timestamp, fin: pd.Timestamp, c15: pd.DataFrame, r151: pd.DataFrame) -> pd.DataFrame:
    """
    Calcule les énergies et les taxes pour une période donnée, sur l'ensemble du périmètre
    """
    historique = lire_flux_c15(c15)
    relevés = lire_flux_r151(r151)
    base = préparer_base_énergies(historique=historique, deb=deb, fin=fin)

    base["Date_Releve_deb"] = base["Date_Releve_deb"].fillna(deb)
    base["Date_Releve_fin"] = base["Date_Releve_fin"].fillna(fin)

    complet = ajouter_relevés(base, relevés)
    return complet

# TODO rename facturation depuis flux ou un truc du genre.
@pa.check_types
def facturation(
        deb: pd.Timestamp, 
        fin: pd.Timestamp, 
        historique: DataFrame[HistoriquePérimètre], 
        relevés: DataFrame[RelevéIndex],
        inclure_jour_fin: bool=False
        ) -> pd.DataFrame:
    """
    Calcule les énergies et les taxes pour une période donnée, sur l'ensemble du périmètre
    """
    historique: DataFrame[HistoriquePérimètre] = extraire_historique_à_date(historique=historique, fin=fin)
    # Base = pour tous les couples (ref, pdl), on a toutes les Entrées/Sorties du périmètre et le relevés associés
    base: DataFrame[BaseCalculEnergies] = préparer_base_énergies(historique=historique, deb=deb, fin=fin)

    # TODO: Gestion des cas spécifiques
    # Pour l'instant, on les met juste de coté.
    cas_spécifiques: DataFrame[ModificationContractuelleImpactante] = extraire_modifications_impactantes(deb, historique)
    # Filtrer la base pour enlever les Ref_Contractuelle présentes dans cas_spécifiques
    if not cas_spécifiques.empty:
        refs_à_exclure = cas_spécifiques["Ref_Situation_Contractuelle"].unique()
        base = base[~base["Ref_Situation_Contractuelle"].isin(refs_à_exclure)]
        print(f"Filtrage de {len(refs_à_exclure)} références contractuelles avec des modifications impactantes")

    # Ajouter les dates de début et de fin pour le cas général (aka les na ici)
    base["Date_Releve_deb"] = base["Date_Releve_deb"].fillna(deb)
    base["Date_Releve_fin"] = base["Date_Releve_fin"].fillna(fin)

    # Ajouter les relevés manquants à la base.
    avec_relevés = base.copy()
    avec_relevés = ajouter_relevés(avec_relevés, relevés, '_deb')
    avec_relevés = ajouter_relevés(avec_relevés, relevés, '_fin')

    énergies = calculer_energies(avec_relevés, inclure_jour_fin)

    régles_turpe = get_applicable_rules(deb, fin)

    turpe = compute_turpe(énergies, régles_turpe)

    # colonnes_triees = sorted(énergies.columns)
    return turpe #.reindex(columns=colonnes_triees)


import numpy as np
import pandas as pd
import pandera.pandas as pa
from pandera.typing import DataFrame, Series
from typing import Annotated, Optional


class FluxR151(pa.DataFrameModel):
    # 📆 Date du relevé
    Date_Releve: Series[Annotated[pd.DatetimeTZDtype, "ns", "Europe/Paris"]] = pa.Field(nullable=False)
    
    # 🔹 Identifiant du Point de Livraison (PDL), aussi appelé Point Réf. Mesures (PRM)
    pdl: Series[str] = pa.Field(nullable=False)

    # 🏢 Références Fournisseur & Distributeur
    Id_Calendrier_Fournisseur: Series[str] = pa.Field(nullable=True)
    Id_Calendrier_Distributeur: Series[str] = pa.Field(nullable=True) 
    Id_Affaire: Series[str] = pa.Field(nullable=True)

    # 📏 Unité de mesure
    Unité: Series[str] = pa.Field(nullable=False, isin=["kWh", "Wh", "MWh"])
    Précision: Series[str] = pa.Field(nullable=False)

    # ⚡ Mesures
    HP: Series[float] = pa.Field(nullable=True, coerce=True)
    HC: Series[float] = pa.Field(nullable=True, coerce=True)
    HCH: Series[float] = pa.Field(nullable=True, coerce=True)
    HPH: Series[float] = pa.Field(nullable=True, coerce=True)
    HPB: Series[float] = pa.Field(nullable=True, coerce=True)
    HCB: Series[float] = pa.Field(nullable=True, coerce=True)
    BASE: Series[float] = pa.Field(nullable=True, coerce=True)

    # 📆 Parser qui converti les Dates en CET "Europe/Paris"
    @pa.dataframe_parser
    def parser_dates(cls, df: DataFrame) -> DataFrame:
        df["Date_Releve"] = (
            pd.to_datetime(df["Date_Releve"], errors="coerce")
            .dt.tz_localize("Europe/Paris")
        )
        return df

    # ⚡ Parser qui converti unités en kWh tout en gardant la précision
    @pa.dataframe_parser
    def parser_unites(cls, df: DataFrame) -> DataFrame:
        cols_index = ["HP", "HC", "BASE", "HCH", "HPH", "HPB", "HCB"]
        df[cols_index] = df[cols_index].apply(pd.to_numeric, errors="coerce")
        
        # Sauvegarde unité originale
        df["Précision"] = df["Unité"]
        
        # Conversion des unités
        
        mask_wh = df["Unité"] == "Wh"
        df.loc[mask_wh, cols_index] /= 1000

        mask_mwh = df["Unité"] == "MWh"
        df.loc[mask_mwh, cols_index] *= 1000

        df["Unité"] = "kWh"
        return df


# Définition du Modèle pour le DataFrame c15
class FluxC15(pa.DataFrameModel):
    # Timestamp
    Date_Evenement: Series[Annotated[pd.DatetimeTZDtype, "ns", "Europe/Paris"]] = pa.Field(nullable=False)

    # Couple d'identifiants
    pdl: Series[str] = pa.Field(nullable=False)
    Ref_Situation_Contractuelle: Series[str] = pa.Field(nullable=False)
    
    # Infos Contractuelles
    Segment_Clientele: Series[str] = pa.Field(nullable=False)
    Etat_Contractuel: Series[str] = pa.Field(nullable=False) # "EN SERVICE", "RESILIE", etc.
    Evenement_Declencheur: Series[str] = pa.Field(nullable=False)  # Ex: "MCT", "MES", "RES"
    Type_Evenement: Series[str] = pa.Field(nullable=False)
    Categorie: Series[str] = pa.Field(nullable=True)

    # Infos calculs tarifs
    Puissance_Souscrite: Series[float] = pa.Field(nullable=False, coerce=True)
    Formule_Tarifaire_Acheminement: Series[str] = pa.Field(nullable=False,)

    # Infos Compteur
    Type_Compteur: Series[str] = pa.Field(nullable=False)
    Num_Compteur: Series[str] = pa.Field(nullable=False)

    # Infos Demande (Optionnel)
    Ref_Demandeur: Series[str] = pa.Field(nullable=True)
    Id_Affaire: Series[str] = pa.Field(nullable=True)
    
    # 📏 Unité de mesure
    Unité: Series[str] = pa.Field(nullable=False, default='kWh')

    Avant_Date_Releve: Series[Annotated[pd.DatetimeTZDtype, "ns", "Europe/Paris"]] = pa.Field(nullable=True)
    Avant_Nature_Index: Series[str] = pa.Field(nullable=True)
    
    # On a parfois deux relevés, dans le cas notamment de changement de calendriers
    Avant_HP: Series[float] = pa.Field(nullable=True, coerce=True)
    Avant_HC: Series[float] = pa.Field(nullable=True, coerce=True)
    Avant_HCH: Series[float] = pa.Field(nullable=True, coerce=True)
    Avant_HPH: Series[float] = pa.Field(nullable=True, coerce=True)
    Avant_HPB: Series[float] = pa.Field(nullable=True, coerce=True)
    Avant_HCB: Series[float] = pa.Field(nullable=True, coerce=True)
    Avant_BASE: Series[float] = pa.Field(nullable=True, coerce=True)

    Après_Date_Releve: Series[Annotated[pd.DatetimeTZDtype, "ns", "Europe/Paris"]] = pa.Field(nullable=True)
    Après_Nature_Index: Series[str] = pa.Field(nullable=True)
    

    Après_HP: Series[float] = pa.Field(nullable=True, coerce=True)
    Après_HC: Series[float] = pa.Field(nullable=True, coerce=True)
    Après_HCH: Series[float] = pa.Field(nullable=True, coerce=True)
    Après_HPH: Series[float] = pa.Field(nullable=True, coerce=True)
    Après_HPB: Series[float] = pa.Field(nullable=True, coerce=True)
    Après_HCB: Series[float] = pa.Field(nullable=True, coerce=True)
    Après_BASE: Series[float] = pa.Field(nullable=True, coerce=True)
    
    @pa.dataframe_parser
    def add_unite(cls, df: DataFrame) -> DataFrame:
        if "Unité" not in df.columns:
            df["Unité"] = "kWh"
        return df

    # 📆 Parser qui converti les Dates en CET "Europe/Paris"
    @pa.dataframe_parser
    def parser_dates(cls, df: DataFrame) -> DataFrame:
        df["Avant_Date_Releve"] = (
            pd.to_datetime(df["Avant_Date_Releve"], utc=True, format="ISO8601")
            .dt.tz_convert("Europe/Paris")
        )
        df["Après_Date_Releve"] = (
            pd.to_datetime(df["Après_Date_Releve"], utc=True, format="ISO8601")
            .dt.tz_convert("Europe/Paris")
        )
        df["Date_Evenement"] = (
            pd.to_datetime(df["Date_Evenement"], utc=True, format="ISO8601")
            .dt.tz_convert("Europe/Paris")
        )
        return df
    
    # ⚡ Parser qui converti unités en kWh tout en gardant la précision
    @pa.dataframe_parser
    def parser_unites(cls, df: DataFrame) -> DataFrame:
        classes_temporelles = ["HP", "HC", "BASE", "HCH", "HPH", "HPB", "HCB"]
        cols_index = (
            ['Avant_'+c for c in classes_temporelles]
            + ['Après_'+c for c in classes_temporelles]
        )
        df[cols_index] = df[cols_index].apply(pd.to_numeric, errors="coerce")
        # Sauvegarde unité originale
        df["Précision"] = df["Unité"]
        
        # Conversion des unités
        mask_wh = df["Unité"] == "Wh"
        df.loc[mask_wh, cols_index] /= 1000

        mask_mwh = df["Unité"] == "MWh"
        df.loc[mask_mwh, cols_index] *= 1000

        df["Unité"] = "kWh"
        return df
    


class FluxF1X(pa.DataFrameModel):
    
    
    # Qu'est-ce qui est facturé ?
    Id_EV: Series[str] = pa.Field(nullable=False)
    Libelle_EV: Series[str] = pa.Field(nullable=False)
    Nature_EV: Optional[Series[str]] = pa.Field(nullable=True) # pas dans le F12
    Date_Debut: Series[Annotated[pd.DatetimeTZDtype, "ns", "Europe/Paris"]] = pa.Field(nullable=False)
    Date_Fin: Series[Annotated[pd.DatetimeTZDtype, "ns", "Europe/Paris"]] = pa.Field(nullable=False)
    
    Unite: Series[str] = pa.Field(nullable=False)
    Quantite: Series[float] = pa.Field(nullable=False, coerce=True)
    

    # 💰 Taxes et Prix
    Prix_Unitaire: Series[float] = pa.Field(nullable=False, coerce=True)
    Montant_HT: Series[float] = pa.Field(nullable=False, coerce=True)
    Taux_TVA_Applicable: Series[str] = pa.Field(nullable=False)

    Formule_Tarifaire_Acheminement: Series[str] = pa.Field(nullable=True)
    

    # 🔹 Identifiant du Point de Livraison (PDL), aussi appelé Point Réf. Mesures (PRM)
    pdl: Series[str] = pa.Field(nullable=False)
    
    # 📆 Infos sur la facture qui contient cette ligne
    Num_Facture: Series[str] = pa.Field(nullable=False)
    Date_Facture: Series[Annotated[pd.DatetimeTZDtype, "ns", "Europe/Paris"]] = pa.Field(nullable=False)
    
    Type_Facturation: Series[str] = pa.Field(nullable=False)

    Source: Series[str] = pa.Field(nullable=False)
    @pa.dataframe_parser
    def parser_source(cls, df: DataFrame) -> DataFrame:
        df["Source"] = "Flux_"+df["Flux"]
        return df
    
    @pa.dataframe_parser
    def parser_colonnes_exclusives(cls, df: DataFrame) -> DataFrame:
        # 🛠️ Transformation spécifique F12
        df_f12 = df[df["Flux"] == "F12"].copy()
        if not df_f12.empty:
            df_f12["Formule_Tarifaire_Acheminement"] = df_f12["Tarif_Souscrit"]
            df_f12["Segment_Clientele"] = df_f12["Code_Segmentation_ERDF"]
            df_f12["Nature_EV"] = None  # Ce champ n'existe pas en F12

        # 🛠️ Transformation spécifique F15
        df_f15 = df[df["Flux"] == "F15"].copy()
        if not df_f15.empty:
            df_f15["Segment_Clientele"] = "C2"  # les F15 sont toujours en C2
            df_f15["CSPE_Applicable"] = True

        # Fusionner les deux sous-DataFrames après transformation
        df_transformed = pd.concat([df_f12, df_f15], ignore_index=True)

        return df_transformed
    
    @pa.parser("Taux_TVA_Applicable")
    def parse_TVA(cls, s: Series[str]) -> Series[str]:
        return pd.to_numeric(s, errors="coerce").fillna(s).astype(str)

    @pa.parser("Nature_EV")
    def parse_nature_EV(cls, s: Series[str]) -> Series[str]:
        
        mapping_catégories = {
            "01": "Acheminement",
            "02": "Prestations et frais",
            "03": "Dispositions contractuelles",
            "04": "Contributions",
        }

        return s.map(mapping_catégories)

    # 📆 Parser qui converti les Dates en CET "Europe/Paris"
    @pa.dataframe_parser
    def parser_dates(cls, df: DataFrame) -> DataFrame:
        for c in ["Date_Debut", "Date_Fin", "Date_Facture"]:
            df[c] = (
                pd.to_datetime(df[c])
                .dt.tz_localize("Europe/Paris")
            )
        return df


import pandas as pd
import pandera.pandas as pa
from pandera.typing import Series, DataFrame
from typing import Annotated, Optional

from icecream import ic
class RelevéIndex(pa.DataFrameModel):
    """
    📌 Modèle Pandera pour les relevés d’index issus de différentes sources.

    Ce modèle permet de valider les relevés de compteurs avec leurs métadonnées.
    """
    # 
    id: Series[str] = pa.Field(nullable=False)

    # 📆 Date du relevé
    Date_Releve: Series[Annotated[pd.DatetimeTZDtype, "ns", "Europe/Paris"]] = pa.Field(nullable=False, coerce=True)
    ordre_index: Series[bool] = pa.Field(default=0)

    # 🔹 Identifiant du Point de Livraison (PDL)
    pdl: Series[str] = pa.Field(nullable=False)
    Ref_Situation_Contractuelle: Optional[Series[str]]

    # 🏢 Références Fournisseur & Distributeur
    Id_Calendrier_Fournisseur: Optional[Series[str]] = pa.Field(nullable=True)  # Peut être absent selon la source
    Id_Calendrier_Distributeur: Series[str] = pa.Field(nullable=True, isin=["DI000001", "DI000002", "DI000003"])
    Id_Affaire: Optional[Series[str]] = pa.Field(nullable=True)  # Référence de la demande associée

    # 
    Source: Series[str] = pa.Field(nullable=False, isin=["flux_R151", "flux_R15", "flux_C15"])

    # 📏 Unité de mesure
    Unité: Series[str] = pa.Field(nullable=False, eq="kWh")
    Précision: Series[str] = pa.Field(nullable=False, isin=["kWh", "Wh", "MWh"])

    # ⚡ Mesures
    HP: Series[float] = pa.Field(nullable=True, coerce=True)
    HC: Series[float] = pa.Field(nullable=True, coerce=True)
    HCH: Series[float] = pa.Field(nullable=True, coerce=True)
    HPH: Series[float] = pa.Field(nullable=True, coerce=True)
    HPB: Series[float] = pa.Field(nullable=True, coerce=True)
    HCB: Series[float] = pa.Field(nullable=True, coerce=True)
    BASE: Series[float] = pa.Field(nullable=True, coerce=True)

    @pa.dataframe_parser
    def parser_ordre_index(cls, df: pd.DataFrame) -> pd.DataFrame:
        df = df.copy()
        if "ordre_index" not in df.columns:
            df["ordre_index"] = False
        else:
            df["ordre_index"] = df["ordre_index"].fillna(False).astype(bool)
        return df
    
    @pa.dataframe_parser
    def parser_id(cls, df: pd.DataFrame) -> pd.DataFrame:
        df = df.copy()
        # Génère un id lisible à partir des colonnes clés
        df["id"] = df[["pdl", "Date_Releve", "Source", "ordre_index"]].astype(str).agg("-".join, axis=1)
        return df
    
    @pa.dataframe_parser
    def parser_colonnes(cls, df: pd.DataFrame) -> pd.DataFrame:
        ordre = list(cls.to_schema().columns.keys())
        df = df.copy()
        df = df[[col for col in ordre if col in df.columns]]
        return df
    
    @pa.dataframe_check
    def verifier_présence_mesures(cls, df: DataFrame) -> bool:
        """Vérifie que les mesures attendues sont présentes selon l'Id_Calendrier_Distributeur."""
        
        # DI000001: BASE non nul
        mask_d1 = df["Id_Calendrier_Distributeur"] == "DI000001"
        base_valide = df.loc[mask_d1, "BASE"].notnull().all()

        # DI000002: HP et HC non nul
        mask_d2 = df["Id_Calendrier_Distributeur"] == "DI000002"
        hp_hc_valide = df.loc[mask_d2, ["HP", "HC"]].notnull().all(axis=1).all()

        # DI000003: HPH, HCH, HPB, HCB non nul
        mask_d3 = df["Id_Calendrier_Distributeur"] == "DI000003"
        hph_hch_hpb_hcb_valide = df.loc[mask_d3, ["HPH", "HCH", "HPB", "HCB"]].notnull().all(axis=1).all()

        # Retourne True si toutes les conditions sont valides
        return base_valide and hp_hc_valide and hph_hch_hpb_hcb_valide
    
class RequêteRelevé(pa.DataFrameModel):
    """
    📌 Modèle Pandera pour les requêtes d'interrogation des relevés d'index.

    Assure que les requêtes sont bien formatées avant d'interroger le DataFrame `RelevéIndex`.
    """
    # 📆 Date du relevé demandée
    Date_Releve: Series[Annotated[pd.DatetimeTZDtype, "ns", "Europe/Paris"]] = pa.Field(nullable=False, coerce=True)

    # 🔹 Identifiant du Point de Livraison (PDL)
    pdl: Series[str] = pa.Field(nullable=False)

    # 


import pandas as pd
def supprimer_colonnes(df: pd.DataFrame)-> pd.DataFrame:
    to_drop = ['turpe_fixe_annuel', 'turpe_fixe_j', 'cg', 'cc', 'b', 'CS_fixe'] + [c for c in df.columns if c.endswith('_rule')]
    to_drop = [c for c in to_drop if c in df.columns]
    return df.drop(columns=to_drop)


def fusion_des_sous_periode(df: pd.DataFrame)-> pd.DataFrame:
    def custom_agg(df):
        agg_dict = {}
        for col in df.columns:
            if pd.api.types.is_numeric_dtype(df[col]):
                agg_dict[col] = lambda x: x.sum(min_count=1)
            else:
                agg_dict[col] = "first"  # Prend la première valeur
        return agg_dict
    # Appliquer le groupby avec la fonction d'agrégation conditionnelle
    return df.groupby("Ref_Situation_Contractuelle").agg(custom_agg(df))

def validation(df: pd.DataFrame)-> pd.DataFrame:
    df['releve_manquant'] = df[['source_releve_fin', 'source_releve_deb']].isna().any(axis=1)
    return df


import pandas as pd
import pandera.pandas as pa
from pandera.typing import Series, DataFrame
from typing import Annotated

# EN QUESTION : En GROS c'est un schéma hybride, composé d'un HistoriquePérimètre + deux RelevésIndex 
# Est-ce qu'on fait un truc d'héritage des modèles ? ou balek ? 
class BaseCalculEnergies(pa.DataFrameModel):
    """
    📌 Modèle Pandera pour la base de calcul des énergies.

    Ce modèle garantit que la base de calcul des énergies est conforme
    et prête à être utilisée pour des calculs plus avancés.
    """
    
    # Sous modèle Pèrimètre : Données contractuelles / Métadonnées
    # Timestamp
    # Date_Evenement: Series[Annotated[pd.DatetimeTZDtype, "ns", "Europe/Paris"]] = pa.Field(nullable=False, coerce=True)

    # Couple d'identifiants
    pdl: Series[str] = pa.Field(nullable=False)
    Ref_Situation_Contractuelle: Series[str] = pa.Field(nullable=False)
    
    # Infos Contractuelles
    Segment_Clientele: Series[str] = pa.Field(nullable=False)
    Etat_Contractuel: Series[str] = pa.Field(nullable=False) # "EN SERVICE", "RESILIE", etc.
    Evenement_Declencheur: Series[str] = pa.Field(nullable=False)  # Ex: "MCT", "MES", "RES"
    Type_Evenement: Series[str] = pa.Field(nullable=False)
    Categorie: Series[str] = pa.Field(nullable=True)

    # Infos calculs tarifs
    Puissance_Souscrite: Series[float] = pa.Field(nullable=False, coerce=True)
    Formule_Tarifaire_Acheminement: Series[str] = pa.Field(nullable=False,)

    # Infos Compteur
    Type_Compteur: Series[str] = pa.Field(nullable=False)
    Num_Compteur: Series[str] = pa.Field(nullable=False)
    
    # Sous modèle RelevéIndex : Relevés début de période
    Date_Releve_deb: Series[Annotated[pd.DatetimeTZDtype, "ns", "Europe/Paris"]] = pa.Field(nullable=True)
    Source_deb: Series[str] = pa.Field(nullable=False, isin=["flux_R151", "flux_R15", "flux_C15"])

    # 📏 Unité de mesure
    Unité_deb: Series[str] = pa.Field(nullable=False, eq="kWh")
    Précision_deb: Series[str] = pa.Field(nullable=False, isin=["kWh", "Wh", "MWh"])

    # ⚡ Mesures
    HP_deb: Series[float] = pa.Field(nullable=True, coerce=True)
    HC_deb: Series[float] = pa.Field(nullable=True, coerce=True)
    HCH_deb: Series[float] = pa.Field(nullable=True, coerce=True)
    HPH_deb: Series[float] = pa.Field(nullable=True, coerce=True)
    HPB_deb: Series[float] = pa.Field(nullable=True, coerce=True)
    HCB_deb: Series[float] = pa.Field(nullable=True, coerce=True)
    BASE_deb: Series[float] = pa.Field(nullable=True, coerce=True)

    # Sous modèle RelevéIndex : Relevés fin de période
    Date_Releve_fin: Series[Annotated[pd.DatetimeTZDtype, "ns", "Europe/Paris"]] = pa.Field(nullable=True)
    Source_fin: Series[str] = pa.Field(nullable=False, isin=["flux_R151", "flux_R15", "flux_C15"])

    # 📏 Unité de mesure
    Unité_fin: Series[str] = pa.Field(nullable=False, eq="kWh")
    Précision_fin: Series[str] = pa.Field(nullable=False, isin=["kWh", "Wh", "MWh"])

    # ⚡ Mesures
    HP_fin: Series[float] = pa.Field(nullable=True, coerce=True)
    HC_fin: Series[float] = pa.Field(nullable=True, coerce=True)
    HCH_fin: Series[float] = pa.Field(nullable=True, coerce=True)
    HPH_fin: Series[float] = pa.Field(nullable=True, coerce=True)
    HPB_fin: Series[float] = pa.Field(nullable=True, coerce=True)
    HCB_fin: Series[float] = pa.Field(nullable=True, coerce=True)
    BASE_fin: Series[float] = pa.Field(nullable=True, coerce=True)


import pandas as pd
import pandera.pandas as pa
from pandera.typing import DataFrame
from babel.dates import format_date

from electricore.core.périmètre.modèles import HistoriquePérimètre, SituationPérimètre, ModificationContractuelleImpactante
from electricore.core.relevés.modèles import RelevéIndex

@pa.check_types
def extraire_situation(date: pd.Timestamp, historique: DataFrame[HistoriquePérimètre]) -> DataFrame[SituationPérimètre]:
    """
    Extrait la situation du périmètre à une date donnée.
    
    Args:
        date (pd.Timestamp): La date de référence.
        historique (pd.DataFrame): L'historique des événements contractuels.

    Returns:
        pd.DataFrame: Une vue du périmètre à `date`, conforme à `SituationPérimètre`.
    """
    return (
        historique[historique["Date_Evenement"] <= date]
        .sort_values(by="Date_Evenement", ascending=False)
        .drop_duplicates(subset=["Ref_Situation_Contractuelle"], keep="first")
    )
@pa.check_types
def extraire_historique_à_date(
    historique: DataFrame[HistoriquePérimètre],
    fin: pd.Timestamp
) -> DataFrame[HistoriquePérimètre]:
    """
    Extrait uniquement les variations (changements contractuels) qui ont eu lieu dans une période donnée.

    Args:
        deb (pd.Timestamp): Début de la période.
        fin (pd.Timestamp): Fin de la période.
        historique (pd.DataFrame): Historique des événements contractuels.

    Returns:
        pd.DataFrame: Un sous-ensemble de l'historique contenant uniquement les variations dans la période.
    """
    return historique[
        (historique["Date_Evenement"] <= fin)
    ].sort_values(by="Date_Evenement", ascending=True)  # Trie par ordre chronologique

@pa.check_types
def extraire_période(
    deb: pd.Timestamp, fin: pd.Timestamp, 
    historique: DataFrame[HistoriquePérimètre]
) -> DataFrame[HistoriquePérimètre]:
    """
    Extrait uniquement les variations (changements contractuels) qui ont eu lieu dans une période donnée.

    Args:
        deb (pd.Timestamp): Début de la période.
        fin (pd.Timestamp): Fin de la période.
        historique (pd.DataFrame): Historique des événements contractuels.

    Returns:
        pd.DataFrame: Un sous-ensemble de l'historique contenant uniquement les variations dans la période.
    """
    return historique[
        (historique["Date_Evenement"] >= deb) & (historique["Date_Evenement"] <= fin)
    ].sort_values(by="Date_Evenement", ascending=True)  # Trie par ordre chronologique

@pa.check_types
def extraite_relevés_entrées(
    historique: DataFrame[HistoriquePérimètre]
) -> DataFrame[RelevéIndex]:
        _événements = ['MES', 'PMES', 'CFNE']
        _colonnes_meta_releve = ['Ref_Situation_Contractuelle', 'pdl', 'Unité', 'Précision', 'Source']
        _colonnes_relevé = ['Id_Calendrier_Distributeur', 'Date_Releve', 'Nature_Index', 'HP', 'HC', 'HCH', 'HPH', 'HPB', 'HCB', 'BASE']
        _colonnes_relevé_après = ['Après_'+c for c in _colonnes_relevé]
        return RelevéIndex.validate(
            historique[historique['Evenement_Declencheur'].isin(_événements)][_colonnes_meta_releve + _colonnes_relevé_après]
            .rename(columns={k: v for k,v in zip(_colonnes_relevé_après, _colonnes_relevé)})
            .dropna(subset=['Date_Releve'])
            )

@pa.check_types
def extraite_relevés_sorties(
    historique: DataFrame[HistoriquePérimètre]
) -> DataFrame[RelevéIndex]:
        _événements = ['RES', 'CFNS']
        _colonnes_meta_releve = ['Ref_Situation_Contractuelle', 'pdl', 'Unité', 'Précision', 'Source']
        _colonnes_relevé = ['Id_Calendrier_Distributeur', 'Date_Releve', 'Nature_Index', 'HP', 'HC', 'HCH', 'HPH', 'HPB', 'HCB', 'BASE']
        _colonnes_relevé_avant = ['Avant_'+c for c in _colonnes_relevé]
        return RelevéIndex.validate(
            historique[historique['Evenement_Declencheur'].isin(_événements)][_colonnes_meta_releve + _colonnes_relevé_avant]
            .rename(columns={k: v for k,v in zip(_colonnes_relevé_avant, _colonnes_relevé)})
            .dropna(subset=['Date_Releve'])
            )

@pa.check_types
def extraire_modifications_impactantes(
    deb: pd.Timestamp,
    historique: DataFrame[HistoriquePérimètre]
) -> DataFrame[ModificationContractuelleImpactante]:
    """
    Détecte les MCT dans une période donnée et renvoie les variations de Puissance_Souscrite
    et Formule_Tarifaire_Acheminement avant et après chaque MCT.

    Args:
        deb (pd.Timestamp): Début de la période.
        historique (pd.DataFrame): Historique des événements contractuels.

    Returns:
        DataFrame[ModificationContractuelleImpactante]: DataFrame contenant les MCT avec les valeurs avant/après.
    """

    # 🔍 Décaler les valeurs pour obtenir les données "avant" AVANT de filtrer
    historique = historique.sort_values(by=["Ref_Situation_Contractuelle", "Date_Evenement"])
    historique["Avant_Puissance_Souscrite"] = historique.groupby("Ref_Situation_Contractuelle")["Puissance_Souscrite"].shift(1)
    historique["Avant_Formule_Tarifaire_Acheminement"] = historique.groupby("Ref_Situation_Contractuelle")["Formule_Tarifaire_Acheminement"].shift(1)


    # 📌 Filtrer uniquement les MCT dans la période donnée
    impacts = (
          historique[
            (historique["Date_Evenement"] >= deb) &
            (historique["Evenement_Declencheur"] == "MCT")]
          .copy()
          .rename(columns={'Puissance_Souscrite': 'Après_Puissance_Souscrite', 'Formule_Tarifaire_Acheminement':'Après_Formule_Tarifaire_Acheminement'})
          .drop(columns=['Segment_Clientele', 'Num_Depannage', 'Categorie', 'Etat_Contractuel', 'Type_Compteur', 'Date_Derniere_Modification_FTA', 'Type_Evenement', 'Ref_Demandeur', 'Id_Affaire'])
    )
    
    # TODO: Prendre en compte plus de cas
    impacts['Impacte_energies'] = (
        impacts["Avant_Id_Calendrier_Distributeur"].notna() & 
        impacts["Après_Id_Calendrier_Distributeur"].notna() & 
        (impacts["Avant_Id_Calendrier_Distributeur"] != impacts["Après_Id_Calendrier_Distributeur"])
    )

    # ➕ Ajout de la colonne de lisibilité du changement
    def generer_resumé(row):
        modifications = []
        if row["Avant_Puissance_Souscrite"] != row["Après_Puissance_Souscrite"]:
            modifications.append(f"P: {row['Avant_Puissance_Souscrite']} → {row['Après_Puissance_Souscrite']}")
        if row["Avant_Formule_Tarifaire_Acheminement"] != row["Après_Formule_Tarifaire_Acheminement"]:
            modifications.append(f"FTA: {row['Avant_Formule_Tarifaire_Acheminement']} → {row['Après_Formule_Tarifaire_Acheminement']}")
        return ", ".join(modifications) if modifications else "Aucun changement"
    
    impacts["Résumé_Modification"] = impacts.apply(generer_resumé, axis=1)

    ordre_colonnes = ModificationContractuelleImpactante.to_schema().columns.keys()
    impacts = impacts[ordre_colonnes]
    
    return impacts

@pa.check_types
def detecter_points_de_rupture(historique: DataFrame[HistoriquePérimètre]) -> DataFrame[HistoriquePérimètre]:
    """
    Enrichit l'historique avec les colonnes d'impact (turpe, énergie, turpe_variable) et un résumé des modifications.
    Toutes les lignes sont conservées.

    Args:
        historique (pd.DataFrame): Historique complet des événements contractuels.

    Returns:
        pd.DataFrame: Historique enrichi avec détection des ruptures et résumé humain.
    """
    index_cols = ['BASE', 'HP', 'HC', 'HPH', 'HCH', 'HPB', 'HCB']

    historique = historique.sort_values(by=["Ref_Situation_Contractuelle", "Date_Evenement"]).copy()
    historique["Avant_Puissance_Souscrite"] = historique.groupby("Ref_Situation_Contractuelle")["Puissance_Souscrite"].shift(1)
    historique["Avant_Formule_Tarifaire_Acheminement"] = historique.groupby("Ref_Situation_Contractuelle")["Formule_Tarifaire_Acheminement"].shift(1)

    impact_turpe_fixe = (
        (historique["Avant_Puissance_Souscrite"].notna() &
         (historique["Avant_Puissance_Souscrite"] != historique["Puissance_Souscrite"])) |
        (historique["Avant_Formule_Tarifaire_Acheminement"].notna() &
         (historique["Avant_Formule_Tarifaire_Acheminement"] != historique["Formule_Tarifaire_Acheminement"]))
    )
    
    changement_calendrier = (
        historique["Avant_Id_Calendrier_Distributeur"].notna() &
        historique["Après_Id_Calendrier_Distributeur"].notna() &
        (historique["Avant_Id_Calendrier_Distributeur"] != historique["Après_Id_Calendrier_Distributeur"])
    )
    
    changement_index = pd.concat([
        (historique[f"Avant_{col}"].notna() &
         historique[f"Après_{col}"].notna() &
         (historique[f"Avant_{col}"] != historique[f"Après_{col}"]))
        for col in index_cols
    ], axis=1).any(axis=1)

    impact_energie = changement_calendrier | changement_index

    impact_turpe_variable = (
      (impact_energie) |
      (historique["Avant_Formule_Tarifaire_Acheminement"].notna() &
         (historique["Avant_Formule_Tarifaire_Acheminement"] != historique["Formule_Tarifaire_Acheminement"]))
    )

    historique["impact_turpe_fixe"] = impact_turpe_fixe
    historique["impact_energie"] = impact_energie
    historique["impact_turpe_variable"] = impact_turpe_variable

    # Forcer les impacts à True pour les événements d’entrée et de sortie
    evenements_entree_sortie = ["CFNE", "MES", "PMES", "CFNS", "RES"]
    mask_entree_sortie = historique["Evenement_Declencheur"].isin(evenements_entree_sortie)

    historique.loc[mask_entree_sortie, ["impact_turpe_fixe", "impact_energie", "impact_turpe_variable"]] = True

    def generer_resume(row):
        modifs = []
        if row["impact_turpe_fixe"]:
            if pd.notna(row.get("Avant_Puissance_Souscrite")) and row["Avant_Puissance_Souscrite"] != row["Puissance_Souscrite"]:
                modifs.append(f"P: {row['Avant_Puissance_Souscrite']} → {row['Puissance_Souscrite']}")
            if pd.notna(row.get("Avant_Formule_Tarifaire_Acheminement")) and row["Avant_Formule_Tarifaire_Acheminement"] != row["Formule_Tarifaire_Acheminement"]:
                modifs.append(f"FTA: {row['Avant_Formule_Tarifaire_Acheminement']} → {row['Formule_Tarifaire_Acheminement']}")
        if row["impact_energie"]:
            modifs.append("rupture index")
        if changement_calendrier.loc[row.name]:
            modifs.append(f"Cal: {row['Avant_Id_Calendrier_Distributeur']} → {row['Après_Id_Calendrier_Distributeur']}")
        return ", ".join(modifs) if modifs else ""

    historique["resume_modification"] = historique.apply(generer_resume, axis=1)

    return historique.reset_index(drop=True)


@pa.check_types
def inserer_evenements_facturation(historique: DataFrame[HistoriquePérimètre]) -> DataFrame[HistoriquePérimètre]:

    tz = "Europe/Paris"

    # Étape 1 : détecter les dates d'entrée et de sortie
    entrees = historique[historique['Evenement_Declencheur'].isin(['CFNE', 'MES', 'PMES'])]
    debuts = entrees.groupby('Ref_Situation_Contractuelle')['Date_Evenement'].min()

    sorties = historique[historique['Evenement_Declencheur'].isin(['RES', 'CFNS'])]
    fins = sorties.groupby('Ref_Situation_Contractuelle')['Date_Evenement'].min()
    today = pd.Timestamp.now(tz=tz).to_period("M").start_time.tz_localize(tz)

    periodes = pd.DataFrame({
        "start": debuts,
        "end": fins
    }).fillna(today)

    # Étape 2 : générer tous les 1ers du mois entre min(start) et max(end)
    min_date = periodes["start"].min()
    max_date = periodes["end"].max()
    all_months = pd.date_range(start=min_date, end=max_date, freq="MS", tz=tz)

    # Étape 3 : associer à chaque ref les mois valides (entre son start et end)
    ref_mois = (
        periodes.reset_index()
        .merge(pd.DataFrame({"Date_Evenement": all_months}), how="cross")
    )
    ref_mois = ref_mois[(ref_mois["Date_Evenement"] >= ref_mois["start"]) & (ref_mois["Date_Evenement"] <= ref_mois["end"])]

    # Étape 4 : créer les événements à ajouter
    evenements = ref_mois.copy()
    evenements["Evenement_Declencheur"] = "FACTURATION"
    evenements["Type_Evenement"] = "artificiel"
    evenements["Source"] = "synthese_mensuelle"
    evenements["resume_modification"] = "Facturation mensuelle"
    evenements["impact_turpe_fixe"] = True
    evenements["impact_energie"] = True
    evenements["impact_turpe_variable"] = True

    evenements = evenements[[
        "Ref_Situation_Contractuelle", "Date_Evenement",
        "Evenement_Declencheur", "Type_Evenement", "Source", "resume_modification",
        "impact_turpe_fixe", "impact_energie", "impact_turpe_variable"
    ]]

    # Étape 5 : concaténer et propager les infos par ffill sur colonnes non-nullables
    fusion = pd.concat([historique, evenements], ignore_index=True).sort_values(
        ["Ref_Situation_Contractuelle", "Date_Evenement"]
    ).reset_index(drop=True)
    
    # Extraire les colonnes non-nullables du modèle Pandera
    colonnes_ffill = [
        name for name, annotation in HistoriquePérimètre.__annotations__.items()
        if name in fusion.columns and HistoriquePérimètre.__fields__[name][1].nullable is False
    ]

    fusion[colonnes_ffill] = (
        fusion.groupby("Ref_Situation_Contractuelle")[colonnes_ffill]
        .ffill()
    )

    # Étape 6 : filtrer uniquement les événements FACTURATION
    ajout = fusion[fusion["Evenement_Declencheur"] == "FACTURATION"]

    # Étape 7 : concat final
    historique_etendu = pd.concat([historique, ajout], ignore_index=True).sort_values(
        ["Ref_Situation_Contractuelle", "Date_Evenement"]
    ).reset_index(drop=True)

    return historique_etendu

@pa.check_types
def extraire_releves_evenements(historique: DataFrame[HistoriquePérimètre]) -> DataFrame[RelevéIndex]:
    """
    Génère des relevés d'index (avant/après) à partir d'un historique enrichi des événements contractuels.

    - Un relevé "avant" (ordre_index=0) est créé à partir des index Avant_*
    - Un relevé "après" (ordre_index=1) est créé à partir des index Après_*
    - La colonne 'ordre_index' permet de trier correctement les relevés successifs.

    Args:
        historique (pd.DataFrame): Historique enrichi (HistoriquePérimètreÉtendu).

    Returns:
        pd.DataFrame: Relevés d’index conformes au modèle RelevéIndex.
    """
    index_cols = ["BASE", "HP", "HC", "HCH", "HPH", "HPB", "HCB", "Id_Calendrier_Distributeur"]
    identifiants = ["pdl"]

    # Créer relevés "avant"
    avant = historique[identifiants + ["Date_Evenement"] + [f"Avant_{col}" for col in index_cols]].copy()
    avant = avant.rename(columns={f"Avant_{col}": col for col in index_cols})
    avant["ordre_index"] = 0

    # Créer relevés "après"
    apres = historique[identifiants + ["Date_Evenement"] + [f"Après_{col}" for col in index_cols]].copy()
    apres = apres.rename(columns={f"Après_{col}": col for col in index_cols})
    apres["ordre_index"] = 1

    # Concaténer
    resultats = pd.concat([avant, apres], ignore_index=True)
    resultats = resultats.dropna(subset=index_cols, how="all")
    resultats["Source"] = "flux_C15"
    resultats["Unité"] = "kWh" 
    resultats["Précision"] = "kWh"

    resultats = resultats.rename(columns={"Date_Evenement": "Date_Releve"})

    # Réordonner selon modèle RelevéIndex (on ne garde que les colonnes présentes)
    colonnes_finales = [col for col in RelevéIndex.to_schema().columns.keys() if col in resultats.columns]
    resultats = resultats[colonnes_finales]

    return resultats


import pandas as pd
import pandera.pandas as pa
from pandera.typing import Series
from typing import Annotated, Optional

class PeriodeAbonnement(pa.DataFrameModel):
    """
    Représente une période homogène de facturation de la part fixe (TURPE)
    pour une situation contractuelle donnée.
    """
    Ref_Situation_Contractuelle: Series[str]
    mois_annee: Series[str]  # ex: "mars 2025"
    periode_debut_lisible: Series[str]  # ex: "1 mars 2025"
    periode_fin_lisible: Series[str]    # ex: "31 mars 2025" ou "en cours"
    Formule_Tarifaire_Acheminement: Series[str]
    Puissance_Souscrite: Series[float]
    nb_jours: Series[int]
    periode_debut: Series[Annotated[pd.DatetimeTZDtype, "ns", "Europe/Paris"]] = pa.Field(nullable=False, coerce=True)

    # Champs de sortie (optionnels)
    turpe_fixe_journalier: Optional[Series[float]] = pa.Field(nullable=True)
    turpe_fixe: Optional[Series[float]] = pa.Field(nullable=True)


# 📌 Main pour tester en dev
import pandas as pd
from electricore.core.périmètre import extraire_situation, variations_mct_dans_periode

if __name__ == "__main__":
    print("🚀 Test de génération de la situation du périmètre")

    # Historique réaliste avec des MCT bien placés
    historique = pd.DataFrame({
        "pdl": [
            "12345", "12345", "12345",  # PDL 12345
            "67890", "67890", "67890",  # PDL 67890
            "11111", "11111"  # PDL 11111
        ],
        "Ref_Situation_Contractuelle": [
            "A1", "A1", "A1",  # Même Ref pour PDL 12345
            "B1", "B1", "B1",  # Même Ref pour PDL 67890
            "C1", "C1"  # Même Ref pour PDL 11111
        ],
        "Date_Evenement": pd.to_datetime([
            "2024-01-01", "2024-02-10", "2024-03-05",  # PDL 12345
            "2024-01-15", "2024-02-20", "2024-03-10",  # PDL 67890
            "2024-02-01", "2024-03-12"  # PDL 11111
        ]).tz_localize("Europe/Paris"),
        "Etat_Contractuel": [
            "EN SERVICE", "EN SERVICE", "RESILIE",  # PDL 12345
            "EN SERVICE", "EN SERVICE", "RESILIE",  # PDL 67890
            "EN SERVICE", "EN SERVICE"  # PDL 11111
        ],
        "Evenement_Declencheur": [
            "MES", "MCT", "RES",  # PDL 12345 (MCT entre MES et RES)
            "MES", "MCT", "RES",  # PDL 67890 (MCT entre MES et RES)
            "MES", "MCT"  # PDL 11111 (MCT en période active)
        ],
        "Type_Evenement": [
            "CONTRAT", "CONTRAT", "CONTRAT",  # PDL 12345
            "CONTRAT", "CONTRAT", "CONTRAT",  # PDL 67890
            "CONTRAT", "CONTRAT"  # PDL 11111
        ],
        "Segment_Clientele": [
            "RES", "RES", "RES",  # PDL 12345
            "RES", "RES", "RES",  # PDL 67890
            "PRO", "PRO"  # PDL 11111
        ],
        "Categorie": [
            "C2", "C2", "C2",  # PDL 12345
            "C2", "C2", "C2",  # PDL 67890
            "C5", "C5"  # PDL 11111
        ],
        "Type_Compteur": [
            "CCB", "CCB", "CCB",  # PDL 12345
            "CCB", "CCB", "CCB",  # PDL 67890
            "CCB", "CCB"  # PDL 11111
        ],
        "Num_Compteur": [
            "C1", "C1", "C1",  # PDL 12345
            "C2", "C2", "C2",  # PDL 67890
            "C3", "C3"  # PDL 11111
        ],
        "Puissance_Souscrite": [
            6, 9, 9,  # Changement à 9 kVA après MCT pour PDL 12345
            12, 15, 15,  # Changement à 15 kVA après MCT pour PDL 67890
            3, 6  # Changement à 6 kVA après MCT pour PDL 11111
        ],
        "Formule_Tarifaire_Acheminement": [
            "BTINFCU4", "BTINFMU4", "BTINFMU4",  # PDL 12345
            "BTINFCUST", "BTINFCU4", "BTINFCU4",  # PDL 67890
            "BTINFCU4", "BTINFMU4"  # PDL 11111
        ],
        "Ref_Demandeur": [
            None, "Dem1", None,  # PDL 12345
            None, "Dem2", None,  # PDL 67890
            None, "Dem3"  # PDL 11111
        ],
        "Id_Affaire": [
            None, "Aff1", None,  # PDL 12345
            None, "Aff2", None,  # PDL 67890
            None, "Aff3"  # PDL 11111
        ]
    })


    # Définir la date de référence
    date_reference = pd.Timestamp("2024-02-15", tz="Europe/Paris")

    # Extraire la situation
    situation = extraire_situation(date_reference, historique)

    # Afficher le résultat
    print("\n📊 Situation du périmètre au", date_reference)
    print(situation)

    # Définir la période d'analyse
    deb = pd.Timestamp("2024-01-01", tz="Europe/Paris")
    fin = pd.Timestamp("2024-03-15", tz="Europe/Paris")

    # Extraire les variations MCT
    variations_mct = variations_mct_dans_periode(deb, fin, historique)

    # Afficher les résultats
    print("\n🔄 Variations MCT dans la période du", deb, "au", fin)
    print(variations_mct)


import pandas as pd
import pandera.pandas as pa
from pandera.typing import Series, DataFrame
from typing import Annotated
"""
| Formulaire SGE / Fiche catalogue / Intervention                                                  | Type_Evenement | Nature_Evenement | Description formulaire / fiche                                                                                           |
| ------------------------------------------------------------------------------------------------ | --------------- | ----------------- | ------------------------------------------------------------------------------------------------------------------------ |
| F100B                                                                                            | CONTRAT         | PMES              | Première mise en service                                                                                                 |
| F120B                                                                                            | CONTRAT         | MES               | Mise en service sur installation existante                                                                               |
| F130                                                                                             | CONTRAT         | CFNS              | Changement de Fournisseur                                                                                                |
| F130                                                                                             | CONTRAT         | CFNE              | Changement de Fournisseur                                                                                                |
| F140B                                                                                            | CONTRAT         | RES               | Résiliation                                                                                                              |
| F140 (RIF)                                                                                       | CONTRAT         | RES               | Résiliation initiative fournisseur                                                                                       |
| F180                                                                                             | CONTRAT         | MCT               | Modification de formule tarifaire ou de puissance d’acheminement                                                         |
| F1020                                                                                            | CONTRAT         | MES               | Mise en service ou rétablissement dans la journée (BT≤36 kVA)                                                            |
| M007                                                                                             | CONTRAT         | AUTRE             | Modification d’information Client                                                                                        |
| F180                                                                                             | TECHNIQUE       | MCF               | Modification de la programmation du calendrier fournisseur                                                               |
| F185                                                                                             | TECHNIQUE       | CMAT              | Modification du dispositif de comptage sans impact sur la formule tarifaire d’acheminement ou sur la puissance souscrite |
| F200B                                                                                            | TECHNIQUE       | COU               | Interventions pour impayé et rétablissement                                                                              |
| F200B                                                                                            | TECHNIQUE       | RET               | Interventions pour impayé et rétablissement                                                                              |
| F1020                                                                                            | TECHNIQUE       | RET               | Mise en service ou rétablissement dans la journée (BT≤36 kVA)                                                            |
| Opération de maintenance sur Compteur ou Disjoncteur                                             | TECHNIQUE       | CMAT              | N/A                                                                                                                      |
| F800B                                                                                            | CONTRAT         | MES               | Mise en service d’un raccordement provisoire pour une durée > 28j                                                        |
| F800B                                                                                            | CONTRAT         | RES               | Résiliation d’un raccordement provisoire pour une durée > 28j                                                            |
| F820                                                                                             | CONTRAT         | MES               | Mise en service d’un raccordement provisoire pour une durée <= 28j                                                       |
| F820                                                                                             | CONTRAT         | RES               | Résiliation d’un raccordement provisoire pour une durée <= 28j                                                           |
| F825                                                                                             | CONTRAT         | MES               | Mise en service d’un raccordement provisoire pour une durée <= 28j (saisie par le Distributeur)                          |
| F825                                                                                             | CONTRAT         | RES               | Résiliation d’un raccordement provisoire pour une durée <= 28j (saisie par le Distributeur)                              |
| F840B                                                                                            | TECHNIQUE       | MDBRA             | Raccordement BT < 36kVA                                                                                                  |
| F860                                                                                             | TECHNIQUE       | CMAT              | Déplacement de comptage / Déplacement ou modification de raccordement                                                    |
| F880                                                                                             | TECHNIQUE       | MDBRA             | Suppression de raccordement                                                                                              |
| F940B                                                                                            | TECHNIQUE       | AUTRE             | Intervention de courte durée                                                                                             |
| M007                                                                                             | CONTRAT         | MDACT             | Modification d’information Client                                                                                        |
| Prestation de visite de contrôle de branchement provisoire (Prolongation Branchement Provisoire) | TECHNIQUE       | AUTRE             | NA                                                                                                                       |
"""

class HistoriquePérimètre(pa.DataFrameModel):
    """
    📌 Modèle Pandera pour l'historique des événements contractuels.
    
    Contient toutes les modifications de périmètre au fil du temps.
    """

    # Timestamp
    Date_Evenement: Series[Annotated[pd.DatetimeTZDtype, "ns", "Europe/Paris"]] = pa.Field(nullable=False, coerce=True)

    # Couple d'identifiants
    pdl: Series[str] = pa.Field(nullable=False)
    Ref_Situation_Contractuelle: Series[str] = pa.Field(nullable=False)
    
    # Infos Contractuelles
    Segment_Clientele: Series[str] = pa.Field(nullable=False)
    Etat_Contractuel: Series[str] = pa.Field(nullable=False) # "EN SERVICE", "RESILIE", etc.
    Evenement_Declencheur: Series[str] = pa.Field(nullable=False)  # Ex: "MCT", "MES", "RES"
    Type_Evenement: Series[str] = pa.Field(nullable=False)
    Categorie: Series[str] = pa.Field(nullable=True)

    # Infos calculs tarifs
    Puissance_Souscrite: Series[float] = pa.Field(nullable=False, coerce=True)
    Formule_Tarifaire_Acheminement: Series[str] = pa.Field(nullable=False,)

    # Infos Compteur
    Type_Compteur: Series[str] = pa.Field(nullable=False)
    Num_Compteur: Series[str] = pa.Field(nullable=False)

    # Infos Demande (Optionnel)
    Ref_Demandeur: Series[str] = pa.Field(nullable=True)
    Id_Affaire: Series[str] = pa.Field(nullable=True)

    # Relève ? On ajoute là on on fait un modèle de relève à part ?

class SituationPérimètre(HistoriquePérimètre):
    """
    📌 Modèle Pandera pour la situation à une date donnée.
    
    Générée à partir de l'historique pour donner un état du périmètre à un instant `t`.
    Chaque `Ref_Situation_Contractuelle` doit être unique.
    """

    @pa.check("Ref_Situation_Contractuelle")
    def unique_ref(cls, series: Series[str]) -> bool:
        """Vérifie que chaque Ref_Situation_Contractuelle est unique dans la situation."""
        return series.is_unique

class ModificationContractuelleImpactante(pa.DataFrameModel):
    """
    📌 Modèle Pandera pour la sortie de `variations_mct_dans_periode`.
    
    Contient les variations de puissance et de tarif après un MCT. (et ptet d'autres, à voir)
    """
    Ref_Situation_Contractuelle: Series[str] = pa.Field(nullable=False)
    pdl: Series[str] = pa.Field(nullable=False)
    Impacte_energies : Series[bool] = pa.Field(nullable=False)
    Résumé_Modification: Series[str] = pa.Field(nullable=False)

    Date_Evenement: Series[Annotated[pd.DatetimeTZDtype, "ns", "Europe/Paris"]] = pa.Field(nullable=False, coerce=True)

    Avant_Puissance_Souscrite: Series[float] = pa.Field(nullable=False, coerce=True)
    Après_Puissance_Souscrite: Series[float] = pa.Field(nullable=False, coerce=True)

    Avant_Formule_Tarifaire_Acheminement: Series[str] = pa.Field(nullable=False)
    Après_Formule_Tarifaire_Acheminement: Series[str] = pa.Field(nullable=False)

    # +relevés avant et Après



import pandas as pd
import pandera.pandas as pa
from pandera.typing import DataFrame
from electricore.core.périmètre.modèles import HistoriquePérimètre
from electricore.inputs.flux.modèles import FluxF1X

from icecream import ic

@pa.check_types()
def lire_flux_f1x(f1x: DataFrame[FluxF1X]): # -> DataFrame[HistoriquePérimètre]:

    df: DataFrame[FluxF1X] = f1x.copy()
    # FluxF15.validate(source)
    df.sort_values(by=["pdl", "Date_Facture"], inplace=True)
    ordre_colonnes = FluxF1X.to_schema().columns.keys()
    df = df[ordre_colonnes]
    return df